/*************************************************
* Base Classes Header File                       *
* (C) 1999-2002 The OpenCL Project               *
*************************************************/

#ifndef OPENCL_BASE_H__
#define OPENCL_BASE_H__

#include <string>
#include <opencl/config.h>
#include <opencl/exceptn.h>
#include <opencl/symkey.h>

namespace OpenCL {

/*************************************************
* Algorithm Base Class                           *
*************************************************/
class Algorithm
   {
   public:
      std::string name() const { return alg_name; }
      virtual void clear() throw() = 0;
      Algorithm(const std::string& n) : alg_name(n) {}
      virtual ~Algorithm() {}
   private:
      const std::string alg_name;
   };

/*************************************************
* Symmetric Algorithm Base Class                 *
*************************************************/
class SymmetricAlgorithm : public Algorithm
   {
   public:
      const u32bit MAXIMUM_KEYLENGTH, MINIMUM_KEYLENGTH, KEYLENGTH_MULTIPLE;
      void set_key(const SymmetricKey&) throw(Invalid_Key_Length);
      void set_key(const byte[], u32bit) throw(Invalid_Key_Length);
      bool valid_keylength(u32bit) const;

      // DEPRECATED FUNCTION
      u32bit default_keylength() const { return MAXIMUM_KEYLENGTH; }

      SymmetricAlgorithm(const std::string&, u32bit, u32bit, u32bit);
      virtual ~SymmetricAlgorithm() {}
   private:
      virtual void key(const byte[], u32bit) = 0;
   };

/*************************************************
* Block Cipher Base Class                        *
*************************************************/
class BlockCipher : public SymmetricAlgorithm
   {
   public:
      const u32bit BLOCK_SIZE;
      void encrypt(const byte in[], byte out[]) const { enc(in, out); }
      void decrypt(const byte in[], byte out[]) const { dec(in, out); }
      void encrypt(byte block[]) const { enc(block, block); }
      void decrypt(byte block[]) const { dec(block, block); }

      // DEPRECATED FUNCTION
      u32bit block_size() const { return BLOCK_SIZE; }

      virtual BlockCipher* clone() const = 0;
      BlockCipher(const std::string&, u32bit, u32bit, u32bit = 0, u32bit = 1);
      virtual ~BlockCipher() {}
   private:
      virtual void enc(const byte[], byte[]) const = 0;
      virtual void dec(const byte[], byte[]) const = 0;
   };

/*************************************************
* Stream Cipher Base Class                       *
*************************************************/
class StreamCipher : public SymmetricAlgorithm
   {
   public:
      void encrypt(const byte i[], byte o[], u32bit len) { cipher(i, o, len); }
      void decrypt(const byte i[], byte o[], u32bit len) { cipher(i, o, len); }
      void encrypt(byte in[], u32bit len) { cipher(in, in, len); }
      void decrypt(byte in[], u32bit len) { cipher(in, in, len); }
      virtual StreamCipher* clone() const = 0;
      StreamCipher(const std::string&, u32bit, u32bit = 0, u32bit = 1);
      virtual ~StreamCipher() {}
   private:
      virtual void cipher(const byte[], byte[], u32bit) = 0;
   };

/*************************************************
* Random Access Stream Cipher Base Class         *
*************************************************/
class RandomAccessStreamCipher : public StreamCipher
   {
   public:
      virtual void seek(u32bit) = 0;
      RandomAccessStreamCipher(const std::string&, u32bit,
                               u32bit = 0, u32bit = 1);
      virtual ~RandomAccessStreamCipher() {}
   };

/*************************************************
* Buffered Computation Base Class                *
*************************************************/
class BufferedComputation : public Algorithm
   {
   public:
      const u32bit OUTPUT_LENGTH;
      void update(const byte[], u32bit);
      void update(const std::string&);
      void final(byte out[]) { final_result(out); }
      SecureVector<byte> final();
      SecureVector<byte> process(const byte[], u32bit);
      SecureVector<byte> process(const std::string&);

      // DEPRECATED FUNCTIONS
      void process(const byte in[], u32bit len, byte out[])
         { update(in, len); final(out); }
      void process(const std::string& in, byte out[])
         { update(in); final(out); }
      u32bit output_length() const { return OUTPUT_LENGTH; }
      // END DEPRECATED FUNCTIONS

      BufferedComputation(const std::string, u32bit);
      virtual ~BufferedComputation() {}
   private:
      virtual void add_data(const byte[], u32bit) = 0;
      virtual void final_result(byte[]) = 0;
   };

/*************************************************
* Hash Function Base Class                       *
*************************************************/
class HashFunction : public BufferedComputation
   {
   public:
      const u32bit HASH_BLOCK_SIZE;
      virtual HashFunction* clone() const = 0;
      HashFunction(const std::string&, u32bit, u32bit = 0);
      virtual ~HashFunction() {}
   };

/*************************************************
* MessageAuthenticationCode Base Class           *
*************************************************/
class MessageAuthenticationCode : public BufferedComputation,
                                  public SymmetricAlgorithm
   {
   public:
      virtual MessageAuthenticationCode* clone() const = 0;
      MessageAuthenticationCode(const std::string&, u32bit, u32bit,
                                u32bit = 0, u32bit = 1);
      virtual ~MessageAuthenticationCode() {}
   };

/*************************************************
* Random Number Generator Base Class             *
*************************************************/
class RandomNumberGenerator : public Algorithm
   {
   public:
      virtual byte random() = 0;
      virtual void randomize(byte[], u32bit);
      u32bit random_u32bit();
      void add_entropy(const byte[], u32bit);
      void add_entropy(class EntropySource&, bool = true);
      RandomNumberGenerator(const std::string& n) : Algorithm(n) {}
      virtual ~RandomNumberGenerator() {}
   private:
      virtual void add_randomness(const byte[], u32bit) throw() = 0;
   };

/*************************************************
* Entropy Source Base Class                      *
*************************************************/
class EntropySource
   {
   public:
      virtual u32bit fast_poll(byte[], u32bit) = 0;
      virtual u32bit slow_poll(byte[], u32bit) = 0;
      virtual ~EntropySource() {}
   };

}

#endif
