/*************************************************
* Luby-Rackoff Header File                       *
* (C) 1999-2002 The OpenCL Project               *
*************************************************/

#ifndef OPENCL_LUBY_RACKOFF_H__
#define OPENCL_LUBY_RACKOFF_H__

#include <opencl/opencl.h>

namespace OpenCL {

template<typename H>
class LubyRackoff : public BlockCipher
   {
   public:
      static std::string name() { return "LubyRackoff<" + H::name() + ">"; }
      static const u32bit BLOCKSIZE = 2*H::HASHLENGTH, KEYLENGTH = 32;

      void clear() throw() { K1.clear(); K2.clear(); }
      BlockCipher* clone() const { return new LubyRackoff<H>; }
      LubyRackoff() : BlockCipher(name(), BLOCKSIZE, 2, 32, 2) {}
   private:
      void enc(const byte[BLOCKSIZE], byte[BLOCKSIZE]) const;
      void dec(const byte[BLOCKSIZE], byte[BLOCKSIZE]) const;
      void key(const byte[], u32bit);
      mutable H hash;
      SecureBuffer<byte, KEYLENGTH/2> K1, K2;
      u32bit keylength;
   };

/*************************************************
* Luby-Rackoff Encryption                        *
*************************************************/
template<typename H>
void LubyRackoff<H>::enc(const byte in[BLOCKSIZE],
                         byte out[BLOCKSIZE]) const
   {
   SecureVector<byte> buffer(hash.OUTPUT_LENGTH);
   hash.update(K1, keylength);
   hash.update(in, BLOCKSIZE/2);
   hash.final(buffer);
   xor_buf(out + BLOCKSIZE/2, in + BLOCKSIZE/2, buffer, BLOCKSIZE/2);

   hash.update(K2, keylength);
   hash.update(out + BLOCKSIZE/2, BLOCKSIZE/2);
   hash.final(buffer);
   xor_buf(out, in, buffer, BLOCKSIZE/2);

   hash.update(K1, keylength);
   hash.update(out, BLOCKSIZE/2);
   hash.final(buffer);
   xor_buf(out + BLOCKSIZE/2, buffer, BLOCKSIZE/2);

   hash.update(K2, keylength);
   hash.update(out + BLOCKSIZE/2, BLOCKSIZE/2);
   hash.final(buffer);
   xor_buf(out, buffer, BLOCKSIZE/2);
   }

/*************************************************
* Luby-Rackoff Decryption                        *
*************************************************/
template<typename H>
void LubyRackoff<H>::dec(const byte in[BLOCKSIZE],
                         byte out[BLOCKSIZE]) const
   {
   SecureBuffer<byte, H::HASHLENGTH> buffer;
   hash.update(K2, keylength);
   hash.update(in + BLOCKSIZE/2, BLOCKSIZE/2);
   hash.final(buffer);
   xor_buf(out, in, buffer, BLOCKSIZE/2);

   hash.update(K1, keylength);
   hash.update(out, BLOCKSIZE/2);
   hash.final(buffer);
   xor_buf(out + BLOCKSIZE/2, in + BLOCKSIZE/2, buffer, BLOCKSIZE/2);

   hash.update(K2, keylength);
   hash.update(out + BLOCKSIZE/2, BLOCKSIZE/2);
   hash.final(buffer);
   xor_buf(out, buffer, BLOCKSIZE/2);

   hash.update(K1, keylength);
   hash.update(out, BLOCKSIZE/2);
   hash.final(buffer);
   xor_buf(out + BLOCKSIZE/2, buffer, BLOCKSIZE/2);
   }

/*************************************************
* Luby-Rackoff Key Schedule                      *
*************************************************/
template<typename H>
void LubyRackoff<H>::key(const byte key[], u32bit length)
   {
   K1.copy(key, length / 2);
   K2.copy(key + length / 2, length / 2);
   keylength = length / 2;
   }

}

#endif
