/*************************************************
* Base Classes Source File                       *
* (C) 1999-2002 The OpenCL Project               *
*************************************************/

#include <opencl/base.h>

namespace OpenCL {

/*************************************************
* SymmetricAlgorithm Constructor                 *
*************************************************/
SymmetricAlgorithm::SymmetricAlgorithm(const std::string& name, u32bit key_min,
                                       u32bit key_max, u32bit key_mod) :
   Algorithm(name),
   MAXIMUM_KEYLENGTH(key_max ? key_max : key_min),
   MINIMUM_KEYLENGTH(key_min),
   KEYLENGTH_MULTIPLE(key_mod)
   {
   }

/*************************************************
* Query if the keylength is valid                *
*************************************************/
bool SymmetricAlgorithm::valid_keylength(u32bit length) const
   {
   return ((length >= MINIMUM_KEYLENGTH) &&
           (length <= MAXIMUM_KEYLENGTH) &&
           (length % KEYLENGTH_MULTIPLE == 0));
   }

/*************************************************
* Set the key                                    *
*************************************************/
void SymmetricAlgorithm::set_key(const SymmetricKey& algo_key)
   throw(Invalid_Key_Length)
   {
   set_key(algo_key, algo_key.length());
   }

/*************************************************
* Set the key                                    *
*************************************************/
void SymmetricAlgorithm::set_key(const byte algo_key[], u32bit length)
   throw(Invalid_Key_Length)
   {
   if(!valid_keylength(length))
      throw Invalid_Key_Length(name(), length);
   key(algo_key, length);
   }

/*************************************************
* BlockCipher Constructor                        *
*************************************************/
BlockCipher::BlockCipher(const std::string& name, u32bit block,
                         u32bit key_min, u32bit key_max, u32bit key_mod) :
   SymmetricAlgorithm(name, key_min, key_max, key_mod),
   BLOCK_SIZE(block)
   {
   }

/*************************************************
* StreamCipher Constructor                       *
*************************************************/
StreamCipher::StreamCipher(const std::string& name, u32bit key_min,
                           u32bit key_max, u32bit key_mod) :
   SymmetricAlgorithm(name, key_min, key_max, key_mod)
   {
   }

/*************************************************
* RandomAccessStreamCipher Constructor           *
*************************************************/
RandomAccessStreamCipher::RandomAccessStreamCipher(const std::string& name,
                                                   u32bit key_min,
                                                   u32bit key_max,
                                                   u32bit key_mod) :
   StreamCipher(name, key_min, key_max, key_mod)
   {
   }

/*************************************************
* BufferedComputation Constructor                *
*************************************************/
BufferedComputation::BufferedComputation(const std::string name, u32bit olen) :
   Algorithm(name), OUTPUT_LENGTH(olen)
   {
   }

/*************************************************
* HashFunction Constructor                       *
*************************************************/
HashFunction::HashFunction(const std::string& name, u32bit hlen, u32bit blen) :
   BufferedComputation(name, hlen), HASH_BLOCK_SIZE(blen)
   {
   }

/*************************************************
* MessageAuthenticationCode Constructor          *
*************************************************/
MessageAuthenticationCode::MessageAuthenticationCode(const std::string& name,
      u32bit mlen, u32bit key_min, u32bit key_max, u32bit key_mod) :
   BufferedComputation(name, mlen),
   SymmetricAlgorithm(name, key_min, key_max, key_mod)
   {
   }

/*************************************************
* Hashing/MACing                                 *
*************************************************/
void BufferedComputation::update(const byte in[], u32bit n)
   {
   add_data(in, n);
   }

void BufferedComputation::update(const std::string& str)
   {
   update((const byte*)str.c_str(), str.size());
   }

SecureVector<byte> BufferedComputation::final()
   {
   SecureVector<byte> output(OUTPUT_LENGTH);
   final_result(output);
   return output;
   }

SecureVector<byte> BufferedComputation::process(const byte in[], u32bit len)
   {
   update(in, len);
   return final();
   }

SecureVector<byte> BufferedComputation::process(const std::string& in)
   {
   update(in);
   return final();
   }

/*************************************************
* Randomize an array of bytes                    *
*************************************************/
void RandomNumberGenerator::randomize(byte out[], u32bit len)
   {
   for(u32bit j = 0; j != len; j++)
      out[j] ^= random();
   }

/*************************************************
* Return a random u32bit                         *
*************************************************/
u32bit RandomNumberGenerator::random_u32bit()
   {
   return ((random() << 24) | (random() << 16) | (random() << 8) | (random()));
   }

/*************************************************
* Add entropy to internal state                  *
*************************************************/
void RandomNumberGenerator::add_entropy(const byte random[], u32bit length)
   {
   add_randomness(random, length);
   }

/*************************************************
* Add entropy to internal state                  *
*************************************************/
void RandomNumberGenerator::add_entropy(EntropySource& source, bool slowpoll)
   {
   SecureVector<byte> entropy(slowpoll ? 128 : 32);
   u32bit returned;
   if(slowpoll) returned = source.slow_poll(entropy, entropy.size());
   else         returned = source.fast_poll(entropy, entropy.size());
   add_entropy(entropy, returned);
   }

}
